@@ -1,3 +1,4 @@
+use std::collections::HashMap;
 use std::net::{IpAddr, Ipv4Addr};
 
 use log::*;
@@ -5,14 +6,14 @@ use serde::{Deserialize, Serialize};
 
 use crate::manager::*;
 
-#[derive(Serialize, Deserialize, Clone, Debug, Default)]
+#[derive(Serialize, Deserialize, Clone, Debug, Default, PartialEq)]
 pub struct PublicKeyWithTime {
     pub key: String, // base64 encoded
     pub priv_key_creation_time: u64,
 }
 
 pub struct PublicPeer {
-    pub public_ip: IpAddr,
+    pub endpoint: String,
     pub wg_port: u16,
     pub admin_port: u16,
     pub wg_ip: Ipv4Addr,
@@ -26,10 +27,11 @@ pub struct StaticConfigurationBuilder {
     wg_name: Option<String>,
     wg_port: Option<u16>,
     admin_port: Option<u16>,
+    subnet: Option<ipnet::Ipv4Net>,
     shared_key: Option<Vec<u8>>,
     my_private_key: Option<String>,
     my_public_key: Option<PublicKeyWithTime>,
-    peers: Vec<PublicPeer>,
+    peers: HashMap<Ipv4Addr, PublicPeer>,
     use_tui: Option<bool>,
     use_existing_interface: Option<bool>,
 }
@@ -61,6 +63,10 @@ impl StaticConfigurationBuilder {
         self.admin_port = Some(port);
         self
     }
+    pub fn subnet(mut self, subnet: ipnet::Ipv4Net) -> Self {
+        self.subnet = Some(subnet);
+        self
+    }
     pub fn shared_key(mut self, shared_key: Vec<u8>) -> Self {
         self.shared_key = Some(shared_key);
         self
@@ -73,7 +79,7 @@ impl StaticConfigurationBuilder {
         self.my_public_key = Some(public_key);
         self
     }
-    pub fn peers(mut self, peers: Vec<PublicPeer>) -> Self {
+    pub fn peers(mut self, peers: HashMap<Ipv4Addr, PublicPeer>) -> Self {
         self.peers = peers;
         self
     }
@@ -86,15 +92,6 @@ impl StaticConfigurationBuilder {
         self
     }
     pub fn build(self) -> StaticConfiguration {
-        let mut myself_as_peer: Option<usize> = None;
-        for (i, peer) in self.peers.iter().enumerate() {
-            if &peer.wg_ip == self.wg_ip.as_ref().unwrap() {
-                debug!("FOUND myself as listener");
-                myself_as_peer = Some(i);
-                break;
-            }
-        }
-
         let peer_cnt = self.peers.len();
         StaticConfiguration {
             name: self.name.unwrap(),
@@ -103,7 +100,7 @@ impl StaticConfigurationBuilder {
             wg_name: self.wg_name.unwrap(),
             wg_port: self.wg_port.unwrap(),
             admin_port: self.admin_port.unwrap(),
-            myself_as_peer,
+            subnet: self.subnet.unwrap(),
             shared_key: self.shared_key.unwrap(),
             my_private_key: self.my_private_key.unwrap(),
             my_public_key: self.my_public_key.unwrap(),
@@ -122,11 +119,11 @@ pub struct StaticConfiguration {
     pub wg_name: String,
     pub wg_port: u16,
     pub admin_port: u16,
-    pub myself_as_peer: Option<usize>,
+    pub subnet: ipnet::Ipv4Net,
     pub shared_key: Vec<u8>,
     pub my_private_key: String,
     pub my_public_key: PublicKeyWithTime,
-    pub peers: Vec<PublicPeer>,
+    pub peers: HashMap<Ipv4Addr, PublicPeer>,
     pub peer_cnt: usize,
     pub use_tui: bool,
     pub use_existing_interface: bool,
@@ -137,14 +134,13 @@ impl StaticConfiguration {
         StaticConfigurationBuilder::new()
     }
     pub fn is_listener(&self) -> bool {
-        self.myself_as_peer.is_some()
+        self.peers.contains_key(&self.wg_ip)
     }
     pub fn as_conf_as_peer(&self, manager: &NetworkManager) -> String {
         let mut lines: Vec<String> = vec![];
         lines.push("[Interface]".to_string());
         lines.push(format!("PrivateKey = {}", self.my_private_key));
-        let port = if let Some(myself) = self.myself_as_peer {
-            let peer = &self.peers[myself];
+        let port = if let Some(peer) = self.peers.get(&self.wg_ip) {
             peer.wg_port
         } else {
             self.wg_port
@@ -152,7 +148,7 @@ impl StaticConfiguration {
         lines.push(format!("ListenPort = {}", port));
         lines.push("".to_string());
 
-        for peer in manager.peer.values() {
+        for peer in manager.peer_iter() {
             lines.push("[Peer]".to_string());
             lines.push(format!("PublicKey = {}", &peer.public_key.key));
             lines.push(format!("AllowedIPs = {}/32", peer.wg_ip));
@@ -160,8 +156,19 @@ impl StaticConfiguration {
             for ip in ips {
                 lines.push(format!("AllowedIPs = {}/32", ip));
             }
-            if let Some(endpoint) = peer.endpoint.as_ref() {
+            if let Some(static_peer) = self.peers.get(&peer.wg_ip) {
+                debug!(target: "configuration", "peer {} uses static endpoint {}", peer.wg_ip, static_peer.endpoint);
+                debug!(target: &peer.wg_ip.to_string(), "use static endpoint {}", static_peer.endpoint);
+                lines.push(format!("EndPoint = {}", static_peer.endpoint));
+            } else if let Some(endpoint) = peer.local_reachable_wg_endpoint.as_ref() {
+                debug!(target: "configuration", "peer {} uses local endpoint {}", peer.wg_ip, endpoint);
+                debug!(target: &peer.wg_ip.to_string(), "use local endpoint {}", endpoint);
                 lines.push(format!("EndPoint = {}", endpoint));
+            } else if let Some(endpoint) = peer.dp_visible_wg_endpoint.as_ref() {
+                debug!(target: "configuration", "peer {} uses visible (NAT) endpoint {}", peer.wg_ip, endpoint);
+                debug!(target: &peer.wg_ip.to_string(), "use visible (NAT) endpoint {}", endpoint);
+                lines.push(format!("EndPoint = {}", endpoint));
+                //                lines.push("PersistentKeepalive = 5".to_string());
             }
             lines.push("".to_string());
         }
@@ -169,11 +176,10 @@ impl StaticConfiguration {
         lines.join("\n")
     }
     pub fn my_admin_port(&self) -> u16 {
-        self.myself_as_peer
-            .map(|i| self.peers[i].admin_port)
-            .unwrap_or(self.admin_port)
-    }
-    pub fn admin_port(&self, peer_index: usize) -> u16 {
-        self.peers[peer_index].admin_port
+        if let Some(peer) = self.peers.get(&self.wg_ip) {
+            peer.admin_port
+        } else {
+            self.admin_port
+        }
     }
 }
