@@ -1,6 +1,6 @@
 use std::fmt;
 use std::net::{IpAddr, Ipv4Addr};
-use std::net::{SocketAddr, ToSocketAddrs, UdpSocket};
+use std::net::{SocketAddr, UdpSocket};
 
 use chacha20poly1305::aead::{Aead, NewAead};
 use chacha20poly1305::{Key, XChaCha20Poly1305, XNonce};
@@ -12,15 +12,26 @@ use crate::configuration::*;
 use crate::error::*;
 use crate::manager::*;
 
+#[derive(Serialize, Deserialize, Debug)]
+pub enum AddressedTo {
+    StaticAddress,
+    LocalAddress,
+    WireguardAddress,
+
+    ReplyFromStaticAddress,
+    ReplyFromLocalAddress,
+    ReplyFromWireguardAddress,
+}
+
 #[derive(Serialize, Deserialize, Debug)]
 pub struct AdvertisementPacket {
+    pub addressed_to: AddressedTo,
     pub public_key: PublicKeyWithTime,
-    pub local_ip_list: Vec<IpAddr>,
     pub local_wg_port: u16,
     pub local_admin_port: u16,
     pub wg_ip: Ipv4Addr,
     pub name: String,
-    pub endpoint: Option<SocketAddr>,
+    pub your_visible_wg_endpoint: Option<SocketAddr>,
     pub routedb_version: usize,
 }
 #[derive(Serialize, Deserialize)]
@@ -30,31 +41,39 @@ pub struct RouteDatabasePacket {
     pub nr_entries: usize,
     pub known_routes: Vec<RouteInfo>,
 }
+#[derive(Serialize, Deserialize, Debug)]
+pub struct LocalContactPacket {
+    pub public_key: PublicKeyWithTime,
+    pub local_ip_list: Vec<IpAddr>,
+    pub local_wg_port: u16,
+    pub local_admin_port: u16,
+    pub my_visible_wg_endpoint: Option<SocketAddr>,
+    pub wg_ip: Ipv4Addr,
+    pub name: String,
+}
 #[derive(Serialize, Deserialize)]
 pub enum UdpPacket {
     Advertisement(AdvertisementPacket),
     RouteDatabaseRequest,
     RouteDatabase(RouteDatabasePacket),
+    LocalContactRequest,
+    LocalContact(LocalContactPacket),
 }
 impl UdpPacket {
     pub fn advertisement_from_config(
         static_config: &StaticConfiguration,
         routedb_version: usize,
+        addressed_to: AddressedTo,
+        to_dynamic_peer: Option<&DynamicPeer>,
     ) -> Self {
-        let endpoint = if static_config.is_listener() {
-            let peer = &static_config.peers[static_config.myself_as_peer.unwrap()];
-            Some(SocketAddr::new(peer.public_ip, peer.wg_port))
-        } else {
-            None
-        };
         UdpPacket::Advertisement(AdvertisementPacket {
+            addressed_to,
             public_key: static_config.my_public_key.clone(),
-            local_ip_list: static_config.ip_list.clone(),
             local_wg_port: static_config.wg_port,
             local_admin_port: static_config.admin_port,
             wg_ip: static_config.wg_ip,
             name: static_config.name.clone(),
-            endpoint,
+            your_visible_wg_endpoint: to_dynamic_peer.and_then(|dp| dp.dp_visible_wg_endpoint),
             routedb_version,
         })
     }
@@ -74,6 +93,23 @@ impl UdpPacket {
             known_routes: known_routes.into_iter().cloned().collect(),
         })
     }
+    pub fn local_contact_request() -> Self {
+        UdpPacket::LocalContactRequest {}
+    }
+    pub fn local_contact_from_config(
+        static_config: &StaticConfiguration,
+        my_visible_wg_endpoint: Option<SocketAddr>,
+    ) -> Self {
+        UdpPacket::LocalContact(LocalContactPacket {
+            public_key: static_config.my_public_key.clone(),
+            local_ip_list: static_config.ip_list.clone(),
+            local_wg_port: static_config.wg_port,
+            local_admin_port: static_config.admin_port,
+            my_visible_wg_endpoint,
+            wg_ip: static_config.wg_ip,
+            name: static_config.name.clone(),
+        })
+    }
 }
 impl fmt::Debug for UdpPacket {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
@@ -81,6 +117,8 @@ impl fmt::Debug for UdpPacket {
             UdpPacket::Advertisement(ad) => ad.fmt(f),
             UdpPacket::RouteDatabaseRequest => f.debug_struct("RouteDatabaseRequest").finish(),
             UdpPacket::RouteDatabase(_) => f.debug_struct("RouteDatabase").finish(),
+            UdpPacket::LocalContactRequest => f.debug_struct("LocalContactRequest").finish(),
+            UdpPacket::LocalContact(_) => f.debug_struct("LocalContact").finish(),
         }
     }
 }
@@ -100,12 +138,17 @@ impl fmt::Debug for UdpPacket {
 pub struct CryptUdp {
     socket: UdpSocket,
     key: Option<[u8; 32]>,
+    udp_send_cnt: usize,
 }
 
 impl CryptUdp {
     pub fn bind(port: u16) -> BoxResult<Self> {
         let socket = UdpSocket::bind(format!("0.0.0.0:{}", port))?;
-        Ok(CryptUdp { socket, key: None })
+        Ok(CryptUdp {
+            socket,
+            key: None,
+            udp_send_cnt: 0,
+        })
     }
     pub fn key(mut self, key: &[u8]) -> BoxResult<Self> {
         if key.len() != 32 {
@@ -121,9 +164,10 @@ impl CryptUdp {
         Ok(CryptUdp {
             socket: self.socket.try_clone()?,
             key: self.key,
+            udp_send_cnt: self.udp_send_cnt,
         })
     }
-    pub fn send_to<T: ToSocketAddrs>(&self, payload: &[u8], addr: T) -> BoxResult<usize> {
+    pub fn send_to(&mut self, payload: &[u8], addr: SocketAddr) -> BoxResult<usize> {
         if let Some(raw_key) = self.key.as_ref() {
             let p = payload.len();
             let padded = ((p + 2 + 7) / 8) * 8; // +2 for 2 Byte length
@@ -150,6 +194,8 @@ impl CryptUdp {
                 .encrypt(nonce, &buf[..])
                 .map_err(|e| format!("{:?}", e))?;
             encrypted.append(&mut nonce_raw.to_vec());
+            self.udp_send_cnt += 1;
+            debug!(target: "udp", "#{}: send {} Bytes to {:?}", self.udp_send_cnt, encrypted.len(), addr);
             Ok(self.socket.send_to(&encrypted, addr)?)
         } else {
             strerror("No encryption key")?
@@ -159,8 +205,10 @@ impl CryptUdp {
         if let Some(raw_key) = self.key.as_ref() {
             let mut enc_buf: Vec<u8> = vec![0; 1500];
             let (length, src_addr) = self.socket.recv_from(&mut enc_buf)?;
+            debug!(target: "udp", "received {} Bytes from {}", length, src_addr);
 
             if length <= 24 {
+                error!(target:"udp", "received buffer too short");
                 strerror("received buffer too short")?;
             }
             let new_length = length - 24;
@@ -171,12 +219,14 @@ impl CryptUdp {
             let cipher = XChaCha20Poly1305::new(key);
             let decrypted = cipher
                 .decrypt(nonce, &enc_buf[..new_length])
-                .map_err(|e| format!("{:?}", e))?;
+                .map_err(|e| format!("Decryption error {:?}", e))?;
 
             if decrypted.len() % 8 != 0 {
+                error!(target:"udp","decrypted buffer is not octet-aligned");
                 strerror("decrypted buffer is not octet-aligned")?;
             }
             if decrypted.len() < 24 {
+                error!(target:"udp","decrypted buffer is too short");
                 strerror("decrypted buffer is too short")?;
             }
 
@@ -192,6 +242,7 @@ impl CryptUdp {
             let crc_received = u64::from_le_bytes(crc_buf);
 
             if crc_received != crc_result {
+                error!(target:"udp","CRC mismatch");
                 strerror("CRC mismatch")?;
             }
 
@@ -205,8 +256,11 @@ impl CryptUdp {
             } else {
                 timestamp - ts_received
             };
-            debug!("UDP TIMESTAMP {}", dt);
+            if dt != 0 {
+                debug!("UDP TIMESTAMP {}", dt);
+            }
             if dt > 10 {
+                error!(target:"udp","time mismatch");
                 strerror("time mismatch")?;
             }
 
@@ -218,6 +272,7 @@ impl CryptUdp {
 
             Ok((p, src_addr))
         } else {
+            error!(target:"udp","No encryption key");
             strerror("No encryption key")?
         }
     }
